from typing import Any, Dict, Optional, Tuple

import matplotlib.pyplot as plt
import meep as mp
import numpy as np
import pandas as pd
import pydantic

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.simulation.gmeep.add_monitors import add_monitors
from gdsfactory.simulation.gmeep.get_material import get_material

mp.verbosity(0)


@pydantic.validate_arguments
def get_transmission_2ports(
    component: Component,
    extend_ports_length: Optional[float] = 4.0,
    layer_core: int = 1,
    layer_source: int = 110,
    layer_monitor1: int = 101,
    layer_monitor2: int = 102,
    layer_simulation_region: int = 2,
    res: int = 20,
    t_clad_bot: float = 1.0,
    t_core: float = 0.22,
    t_clad_top: float = 1.0,
    dpml: int = 1,
    clad_material: str = "SiO2",
    core_material: str = "Si",
    is_3d: bool = False,
    run: bool = True,
    wavelength_min: float = 1.5,
    wavelength_max: float = 1.6,
    wavelength_points: int = 50,
    field_monitor_point: Tuple[int, int, int] = (0, 0, 0),
    dfcen: float = 0.2,
) -> Dict[str, Any]:
    """Returns dict with Sparameters for a 2port gf.component

    requires source and  port monitors in the GDS

    based on meep directional coupler example
    https://meep.readthedocs.io/en/latest/Python_Tutorials/GDSII_Import/

    https://support.lumerical.com/hc/en-us/articles/360042095873-Metamaterial-S-parameter-extraction

    Args:
        component: gf.Component
        extend_ports_function: function to extend ports beyond the PML
        layer_core: GDS layer for the Component material
        layer_source: for the source monitor
        layer_monitor1: monitor layer for port 1
        layer_monitor2: monitor layer for port 2
        layer_simulation_region: for simulation region
        res: resolution (pixels/um) For example: (10: 100nm step size)
        t_clad_bot: thickness for cladding below core
        t_core: thickness of the core material
        t_clad_top: thickness for cladding above core
        dpml: PML thickness (um)
        clad_material: material for cladding
        core_material: material for core
        is_3d: if True runs in 3D
        run: if True runs simulation, False only build simulation
        wavelengths: iterable of wavelengths to simulate
        field_monitor_point: monitors field and stops simulation after field decays by 1e-9
        dfcen: delta frequency

    Returns:
        Dict:
            sim: simulation object

    Make sure you visualize the simulation with before you run

    .. code::

        import gdsfactory as gf
        import gdsfactory.simulation.meep as gm

        component = gf.components.bend_circular()
        margin = 2
        cm = gm.add_monitors(component)
        cm.show()

    """
    clad_material = get_material(name=clad_material)
    core_material = get_material(name=core_material)

    assert isinstance(
        component, Component
    ), f"component needs to be a Component, got Type {type(component)}"
    if extend_ports_length:
        component = gf.components.extension.extend_ports(
            component=component, length=extend_ports_length, centered=True
        )
    component.flatten()
    gdspath = component.write_gds()
    gdspath = str(gdspath)

    wavelengths = np.linspace(wavelength_min, wavelength_max, wavelength_points)
    freqs = 1 / wavelengths
    fcen = np.mean(freqs)
    frequency_width = dfcen * fcen
    cell_thickness = dpml + t_clad_bot + t_core + t_clad_top + dpml

    cell_zmax = 0.5 * cell_thickness if is_3d else 0
    cell_zmin = -0.5 * cell_thickness if is_3d else 0

    core_zmax = 0.5 * t_core if is_3d else 10
    core_zmin = -0.5 * t_core if is_3d else -10

    geometry = mp.get_GDSII_prisms(
        core_material, gdspath, layer_core, core_zmin, core_zmax
    )
    cell = mp.GDSII_vol(gdspath, layer_core, cell_zmin, cell_zmax)
    sim_region = mp.GDSII_vol(gdspath, layer_simulation_region, cell_zmin, cell_zmax)

    cell.size = mp.Vector3(
        sim_region.size[0] + 2 * dpml, sim_region.size[1] + 2 * dpml, sim_region.size[2]
    )
    cell_size = cell.size

    zsim = t_core + t_clad_top + t_clad_bot + 2 * dpml
    m_zmin = -zsim / 2
    m_zmax = +zsim / 2
    src_vol = mp.GDSII_vol(gdspath, layer_source, m_zmin, m_zmax)

    sources = [
        mp.EigenModeSource(
            src=mp.GaussianSource(fcen, fwidth=frequency_width),
            size=src_vol.size,
            center=src_vol.center,
            eig_band=1,
            eig_parity=mp.NO_PARITY if is_3d else mp.EVEN_Y + mp.ODD_Z,
            eig_match_freq=True,
        )
    ]

    sim = mp.Simulation(
        resolution=res,
        cell_size=cell_size,
        boundary_layers=[mp.PML(dpml)],
        sources=sources,
        geometry=geometry,
        default_material=clad_material,
    )
    sim_settings = dict(
        resolution=res,
        cell_size=cell_size,
        fcen=fcen,
        field_monitor_point=field_monitor_point,
        layer_core=layer_core,
        t_clad_bot=t_clad_bot,
        t_core=t_core,
        t_clad_top=t_clad_top,
        is_3d=is_3d,
        dmp=dpml,
    )

    m1_vol = mp.GDSII_vol(gdspath, layer_monitor1, m_zmin, m_zmax)
    m2_vol = mp.GDSII_vol(gdspath, layer_monitor2, m_zmin, m_zmax)
    m1 = sim.add_mode_monitor(
        freqs,
        mp.ModeRegion(center=m1_vol.center, size=m1_vol.size),
    )
    m1.z = 0
    m2 = sim.add_mode_monitor(
        freqs,
        mp.ModeRegion(center=m2_vol.center, size=m2_vol.size),
    )
    m2.z = 0

    # if 0:
    #     ''' Useful for debugging.  '''
    #     sim.run(until=50)
    #     sim.plot2D(fields=mp.Ez)
    #     plt.show()
    #     quit()

    r = dict(sim=sim, cell_size=cell_size, sim_settings=sim_settings)

    if run:
        sim.run(
            until_after_sources=mp.stop_when_fields_decayed(
                dt=50, c=mp.Ez, pt=field_monitor_point, decay_by=1e-9
            )
        )

        # call this function every 50 time spes
        # look at simulation and measure component that we want to measure (Ez component)
        # when field_monitor_point decays below a certain 1e-9 field threshold

        # Calculate the mode overlaps
        m1_results = sim.get_eigenmode_coefficients(m1, [1]).alpha
        m2_results = sim.get_eigenmode_coefficients(m2, [1]).alpha

        # Parse out the overlaps
        a1 = m1_results[:, :, 0]  # forward wave
        b1 = m1_results[:, :, 1]  # backward wave
        a2 = m2_results[:, :, 0]  # forward wave
        # b2 = m2_results[:, :, 1]  # backward wave

        # Calculate the actual scattering parameters from the overlaps
        s11 = np.squeeze(b1 / a1)
        s12 = np.squeeze(a2 / a1)
        s22 = s11.copy()
        s21 = s12.copy()

        # s22 and s21 requires another simulation, with the source on the other port
        # Luckily, if the device is symmetric, we can assume that s22=s11 and s21=s12.

        # visualize results
        plt.figure()
        plt.plot(
            wavelengths,
            10 * np.log10(np.abs(s11) ** 2),
            "-o",
            label="Reflection",
        )
        plt.plot(
            wavelengths,
            10 * np.log10(np.abs(s12) ** 2),
            "-o",
            label="Transmission",
        )
        plt.ylabel("Power (dB)")
        plt.xlabel(r"Wavelength ($\mu$m)")
        plt.legend()
        plt.grid(True)

        r.update(dict(s11=s11, s12=s12, s21=s21, s22=s22, wavelengths=wavelengths))
        keys = [key for key in r.keys() if key.startswith("S")]
        s = {f"{key}a": list(np.unwrap(np.angle(r[key].flatten()))) for key in keys}
        s_mod = {f"{key}m": list(np.abs(r[key].flatten())) for key in keys}
        s.update(**s_mod)
        s = pd.DataFrame(s)
    return r


def plot2D(results_dict, z=0):
    """Plot a 2D cut of your simulation."""
    sim = results_dict["sim"]
    cell_size = results_dict["cell_size"]
    cell_size.z = 0
    sim.plot2D(
        output_plane=mp.Volume(center=mp.Vector3(), size=cell_size),
        fields=mp.Ez,
        field_parameters={"interpolation": "spline36", "cmap": "RdBu"},
    )


def plot3D(results_dict):
    """Plots 3D simulation in Mayavi."""
    sim = results_dict["sim"]
    sim.plot3D()


def test_waveguide_2D() -> None:
    """Ensure >99% transmission (S21) at 1550nm."""
    c = gf.components.straight(length=2)
    cm = add_monitors(component=c)
    # gf.show(cm)

    r = get_transmission_2ports(cm, is_3d=False, run=True)
    assert 0.99 < np.mean(abs(r["s21"])) < 1.01
    assert 0 < np.mean(abs(r["s11"])) < 0.2


# def test_waveguide_3D() -> None:
#     """Ensure >99% transmission (S21) at 1550nm."""
#     c = gf.components.straight(length=2)
#     cm = add_monitors(component=c)
#     gf.show(cm)

#     r = get_transmission_2ports(cm, is_3d=True, run=True, res=10)
#     assert 0.99 < np.mean(abs(r["s21"])) < 1.01
#     assert 0 < np.mean(abs(r["s11"])) < 0.2


def test_bend_2D():
    """Ensure >99% transmission (S21) at 1550nm."""
    c = gf.components.bend_circular(radius=5)
    cm = add_monitors(component=c)
    # gf.show(cm)

    r = get_transmission_2ports(cm, is_3d=False, run=True)
    assert 0.97 < np.mean(abs(r["s21"])) < 1.01
    assert 0 < np.mean(abs(r["s11"])) < 0.2


if __name__ == "__main__":
    c = gf.components.straight(length=2)
    cm = add_monitors(component=c)
    gf.show(cm)

    r = get_transmission_2ports(cm, run=True)
    print(r)

    # sim = r["sim"]
    # plt.show()



import gdsfactory as gf
from gdsfactory.tech import LAYER


@gf.cell
def test_path_extrude_multiple_ports() -> gf.Component:
    X = gf.CrossSection()
    X.add(width=0.5, offset=0, layer=LAYER.SLAB90, ports=["o1", "o2"])
    X.add(
        width=2.0,
        offset=-4,
        layer=LAYER.HEATER,
        ports=["e1", "e2"],
        port_types=("electrical", "electrical"),
    )
    P = gf.path.straight(npoints=100, length=10)
    c = gf.path.extrude(P, X)
    assert c.ports["e1"].port_type == "electrical"
    assert c.ports["e2"].port_type == "electrical"
    assert c.ports["o1"].port_type == "optical"
    assert c.ports["o2"].port_type == "optical"
    return c


if __name__ == "__main__":
    c = test_path_extrude_multiple_ports()
    c.show()



from typing import Tuple

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.fiber import circle


@gf.cell
def fiber_array(
    n: int = 8,
    pitch: float = 127.0,
    core_diameter: float = 10,
    cladding_diameter: float = 125,
    layer_core: Tuple[int, int] = gf.LAYER.WG,
    layer_cladding: Tuple[int, int] = gf.LAYER.WGCLAD,
) -> Component:
    """Returns a fiber array

    .. code::

        pitch
         <->
        _________
       |         | lid
       | o o o o |
       |         | base
       |_________|
          length

    """
    c = Component()

    for i in range(n):
        core = c.add_ref(circle(radius=core_diameter / 2, layer=layer_core))
        cladding = c.add_ref(circle(radius=cladding_diameter / 2, layer=layer_cladding))
        core.movex(i * pitch)
        cladding.movex(i * pitch)
        c.add_port(name=f"F{i}", width=core_diameter, orientation=0)

    return c


if __name__ == "__main__":
    c = fiber_array()
    c.show(show_ports=True)



"""
[tidy3D is a FDTD web based software](https://simulation.cloud/)

"""

from gdsfactory.simulation.tidy3d import materials
from gdsfactory.simulation.tidy3d.get_coupling import get_coupling
from gdsfactory.simulation.tidy3d.get_simulation import get_simulation, plot_simulation
from gdsfactory.simulation.tidy3d.get_sparameters import get_sparameters
from gdsfactory.simulation.tidy3d.run_simulation import run_simulation

__version__ = "0.0.1"
__all__ = [
    "plot_simulation",
    "get_coupling",
    "get_sparameters",
    "get_simulation",
    "run_simulation",
    "materials",
]



from typing import List, Tuple

import gdsfactory as gf
from gdsfactory import components as pc
from gdsfactory.component import Component


@gf.cell
def litho_steps(
    line_widths: List[float] = (1.0, 2.0, 4.0, 8.0, 16.0),
    line_spacing: float = 10.0,
    height: float = 100.0,
    layer: Tuple[int, int] = gf.LAYER.WG,
) -> Component:
    """Produces a positive + negative tone linewidth test, used for
    lithography resolution test patterning
    adapted from phidl

    Args:
        line_widths:
        line_spacing:
        height:
        layer:

    """
    D = gf.Component()

    height = height / 2
    T1 = pc.text(
        text="%s" % str(line_widths[-1]), size=height, justify="center", layer=layer
    )
    D.add_ref(T1).rotate(90).movex(-height / 10)
    R1 = pc.rectangle(size=(line_spacing, height), layer=layer)
    D.add_ref(R1).movey(-height)
    count = 0
    for i in reversed(line_widths):
        count += line_spacing + i
        R2 = pc.rectangle(size=(i, height), layer=layer)
        D.add_ref(R1).movex(count).movey(-height)
        D.add_ref(R2).movex(count - i)

    return D


if __name__ == "__main__":
    c = litho_steps()
    c.show()



"""when you add references you have to make sure they have unique names.

the cell decorator gives unique names to components that depend on their parameters
"""


from typing import Union

import gdsfactory as gf
from gdsfactory.component import Component


@gf.cell
def straight(width: Union[float, int] = 10, height: int = 1) -> Component:
    """Returns straight with automatic name."""
    wg = gf.Component("straight")
    wg.add_polygon([(0, 0), (width, 0), (width, height), (0, height)])
    wg.add_port(name="wgport1", midpoint=[0, height / 2], width=height, orientation=180)
    wg.add_port(
        name="wgport2", midpoint=[width, height / 2], width=height, orientation=0
    )
    return wg


def test_autoname() -> None:
    c = straight()
    assert c.name.split("_")[0] == "straight"


if __name__ == "__main__":
    c1 = straight()
    print(c1)

    c2 = straight(width=0.5)
    print(c2)

    c3 = straight(0.5)
    assert c2.name == c3.name



import numpy as np
import tidy3d as td

from gdsfactory.simulation.tidy3d.run_simulation import run_simulation


def get_sparameters(sim: td.Simulation) -> np.ndarray:
    """Adapted from tidy3d examples.

    Returns full Smatrix for a component

    https://support.lumerical.com/hc/en-us/articles/360042095873-Metamaterial-S-parameter-extraction
    """

    sim = run_simulation(sim).result()

    def get_amplitude(monitor):
        f, b = sim.data(monitor)["mode_amps"]
        return np.squeeze(f), np.squeeze(b)

    monitors = sim.monitors

    n = len(monitors) - 1
    S = np.zeros((n, n), dtype=np.complex128)

    # for i, monitor_i in enumerate(monitors):
    #     for j, monitor_j in enumerate(monitors):
    #         if i > 0 and j > 0:
    #             if monitor_i.name.startswith("W"):
    #                 ai, bi = get_amplitude(monitor_i)
    #             else:
    #                 bi, ai = get_amplitude(monitor_i)
    #             if monitor_j.name.startswith("W"):
    #                 aj, bj = get_amplitude(monitor_j)
    #             else:
    #                 bj, aj = get_amplitude(monitor_j)
    #             S[i - i, j - 1] = bi / aj

    if len(monitors) == 5:
        _, incident, reflect, top, bot = monitors
        S[0, 0] = get_amplitude(incident)[-1]
        S[1, 0] = get_amplitude(reflect)[-1]
        S[0, 1] = get_amplitude(top)[0]
        S[1, 1] = get_amplitude(bot)[0]

    elif len(monitors) == 3:
        _, incident, reflect = monitors
        S[0, 0] = S[1, 1] = get_amplitude(incident)[-1]
        S[1, 0] = S[0, 1] = get_amplitude(reflect)[-1]
    return S


if __name__ == "__main__":
    import gdsfactory as gf
    import gdsfactory.simulation.tidy3d as gm

    c = gf.components.straight(length=2)
    sim = gm.get_simulation(c)
    # s = get_sparameters(sim)
    # print(s)

    sim = run_simulation(sim).result()

    def get_amplitude(monitor):
        f, b = sim.data(monitor)["mode_amps"]
        # return f, b
        return np.squeeze(f), np.squeeze(b)

    monitors = sim.monitors

    n = len(monitors) - 1
    S = np.zeros((n, n), dtype=np.complex128)

    # for i, monitor_i in enumerate(monitors):
    #     for j, monitor_j in enumerate(monitors):
    #         if i > 0 and j > 0:
    #             # ai, bi = get_amplitude(monitor_i)
    #             # aj, bj = get_amplitude(monitor_j)
    #             # S[i-1, j-1] = bi / aj
    #             if monitor_i.name.startswith("W"):
    #                 ai, bi = get_amplitude(monitor_i)
    #             else:
    #                 bi, ai = get_amplitude(monitor_i)
    #             if monitor_j.name.startswith("W"):
    #                 aj, bj = get_amplitude(monitor_j)
    #             else:
    #                 bj, aj = get_amplitude(monitor_j)
    #             S[i - 1, j - 1] = bi / aj

    _, incident, reflect = monitors
    a1, b1 = get_amplitude(incident)
    b2, a2 = get_amplitude(reflect)
    S[0, 0] = S[1, 1] = b1 / a1  # S11
    S[1, 0] = S[0, 1] = b2 / a1  # S12



import pathlib
from functools import partial

import matplotlib.pyplot as plt
import meep as mp
import numpy as np
import pandas as pd
import pydantic

from gdsfactory.simulation.modes.find_modes import find_modes
from gdsfactory.simulation.modes.get_mode_solver_rib import get_mode_solver_rib
from gdsfactory.simulation.modes.types import ModeSolverFactory
from gdsfactory.types import Optional, PathType


@pydantic.validate_arguments
def find_neff_vs_width(
    w1: float = 0.2,
    w2: float = 1.0,
    steps: int = 12,
    mode_solver: ModeSolverFactory = get_mode_solver_rib,
    nmodes: int = 4,
    wavelength: float = 1.55,
    parity=mp.NO_PARITY,
    filepath: Optional[PathType] = None,
    overwrite: bool = False,
    **kwargs
) -> pd.DataFrame:

    if filepath and not overwrite and pathlib.Path(filepath).exists():
        return pd.read_csv(filepath)

    width = np.linspace(w1, w2, steps)
    neff = {}
    for mode_number in range(1, nmodes + 1):
        neff[mode_number] = []

    for wg_width in width:
        mode_solver = partial(mode_solver, wg_width=wg_width, nmodes=nmodes, **kwargs)
        modes = find_modes(mode_solver, wavelength=wavelength, parity=parity)
        for mode_number in range(1, nmodes + 1):
            mode = modes[mode_number]
            neff[mode_number].append(mode.neff)

    df = pd.DataFrame(neff)
    df["width"] = width
    if filepath:
        filepath = pathlib.Path(filepath)
        dirpath = filepath.parent
        dirpath.mkdir(exist_ok=True, parents=True)
        df.to_csv(filepath, index=False)
    return df


def plot_neff_vs_width(df: pd.DataFrame, **kwargs):
    width = df.width
    for mode_number, neff in df.items():
        if mode_number != "width":
            plt.plot(width, neff, ".-", label=str(mode_number))

    plt.legend(**kwargs)
    plt.xlabel("width (um)")
    plt.ylabel("neff")


if __name__ == "__main__":
    df = find_neff_vs_width(steps=3, filepath="neff_vs_width.csv")
    plot_neff_vs_width(df)
    plt.show()

    # w1: float = 0.5
    # w2: float = 1.0
    # steps: int = 3
    # mode_solver = get_mode_solver_rib
    # nmodes: int = 4
    # wavelength: float = 1.55
    # parity = mp.NO_PARITY
    # width = np.linspace(w1, w2, steps)

    # neff = {}

    # for mode_number in range(1, nmodes + 1):
    #     neff[mode_number] = []

    # for wg_width in width:
    #     mode_solver = partial(mode_solver, wg_width=wg_width, nmodes=nmodes)
    #     modes = find_modes(mode_solver, wavelength=wavelength, parity=parity)
    #     for mode_number in range(1, nmodes + 1):
    #         mode = modes[mode_number]
    #         neff[mode_number].append(mode.neff)

    # s = WidthSweep(width=list(width), neff=neff)



"""gdsfactory interface to simulations."""

from gdsfactory.simulation import plot
from gdsfactory.simulation.get_sparameters_path import get_sparameters_path
from gdsfactory.simulation.read import (
    read_sparameters_lumerical,
    read_sparameters_pandas,
)
from gdsfactory.simulation.write_sparameters_components_lumerical import (
    write_sparameters_components_lumerical,
)
from gdsfactory.simulation.write_sparameters_lumerical import (
    write_sparameters_lumerical,
)

__all__ = [
    "get_sparameters_path",
    "plot",
    "read_sparameters_pandas",
    "read_sparameters_lumerical",
    "write_sparameters_lumerical",
    "write_sparameters_components_lumerical",
]



"""A Design Of Experiment (DOE) or sweep
changes one or several parameters.
"""

import itertools as it
from io import StringIO
from pathlib import Path
from typing import Any, Dict, List, Union

from omegaconf import OmegaConf

from gdsfactory.config import CONFIG

_sample_yaml_code = """
does:
    cutback_taper_te_400:
      component: cutback_taper_te
      settings:
        wg_width: [0.4]
        length: [1.5]
        n_devices_target: [30, 60, 120]

"""


def read_sweep(filepath: Union[Path, StringIO]) -> Dict[str, Any]:
    """Return dictionary with the information loaded from does.yml

    Args:
        filepath: yaml file describing does

    Returns:
        a dictionary of DOEs with:
        {
            doe_name1: [(component_factory_name, parameters), ...]
            doe_name2: [(component_factory_name, parameters), ...]
            ...
        }

    .. code:: yaml

        mmi1x2_gap:
            component: mmi1x2
            gap: [0.5, 0.6]
            length: 10

        mmi1x2:
            component: mmi1x2
            length: [11, 12]
            gap: [0.2, 0.3]
            do_permutation: False

    """
    does = {}
    input_does = OmegaConf.load(filepath)
    input_does = OmegaConf.to_container(input_does)

    for doe_name, doe in input_does.items():
        if doe_name == "mask":
            continue
        if doe.get("type", "") == "template":
            continue

        if "do_permutation" in doe:
            do_permutation = doe.pop("do_permutation")
        else:
            do_permutation = True
        if not doe.get("settings"):
            raise ValueError(f"Error, missing settings: for {doe_name}")

        doe_settings = doe.pop("settings", "")
        if doe_settings:
            doe["settings"] = get_settings_list(do_permutation, **doe_settings)
        else:
            raise ValueError(
                f"DOE {doe_name} is not a dictionary",
                f"\n\t got: {doe}\n\t sample: {_sample_yaml_code}",
            )

        does[doe_name] = doe

    return does


def get_settings_list(do_permutations: bool = True, **kwargs) -> List[Dict[str, Any]]:
    """Return a list of settings

    Args:
        do_permutations: if False, will only zip the values passed for each parameter
            True creates all combinations
        kwargs: Keyword arguments with a list or tuple of desired values to sweep

    Usage:

        import gdsfactory as gf

        gf.sweep.read_sweep.get_settings_list(length=[30, 40])  # adds different lengths
        gf.sweep.read_sweep.get_settings_list(length=[30, 40], width=[4, 8])
        gf.sweep.read_sweep.get_settings_list(length=[30, 40], width=[4, 8])
        # if do_permutations=True, does all combinations (L30W4, L30W8, L40W4, L40W8)
        # if do_permutations=False, zips arguments (L30W4, L40W8)


    add variations of self.baseclass in self.components
    get arguments from default_args and then update them from kwargs
    updates default_args with kwargs
    self.settings lists all the variations
    """

    # Deal with empty parameter case
    if kwargs == {}:
        return [dict()]

    # Accept both values or lists
    for key, value in list(kwargs.items()):
        if not isinstance(value, list):
            kwargs[key] = [value]

    if do_permutations:
        keys, list_values = list(zip(*list(kwargs.items())))
        settings = [dict(list(zip(keys, perms))) for perms in it.product(*list_values)]
    else:
        keys, list_values = list(zip(*list(kwargs.items())))
        settings = [dict(list(zip(keys, values))) for values in zip(*list_values)]

    return settings


def test_read_sweep() -> Dict[str, Any]:
    filepath = CONFIG["samples_path"] / "mask" / "does.yml"
    does = read_sweep(filepath)
    assert len(does) == 4
    return does


if __name__ == "__main__":
    test_read_sweep()
    # from pprint import pprint

    # does_path = CONFIG["samples_path"] / "mask" / "does.yml"
    # pprint(load_does(does_path))



"""From phidl

"""
from typing import Tuple

import numpy as np
from phidl.device_layout import Group

from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.text_rectangular import text_rectangular
from gdsfactory.difftest import difftest
from gdsfactory.types import ComponentFactory, ComponentOrFactory, Float2


@cell
def grid(
    components: Tuple[ComponentOrFactory, ...],
    spacing: Tuple[float, float] = (5.0, 5.0),
    separation: bool = True,
    shape: Tuple[int, int] = None,
    align_x: str = "x",
    align_y: str = "y",
    edge_x: str = "x",
    edge_y: str = "ymax",
    rotation: int = 0,
    h_mirror: bool = False,
    v_mirror: bool = False,
) -> Component:
    """Returns a component with a 1D or 2D grid of components

    Adapted from phid.geometry

    Args:
        components: Iterable to be placed onto a grid. (can be 1D or 2D)
        spacing: between adjacent elements on the grid, can be a tuple for
          different distances in height and width.
        separation: If True, guarantees elements are speparated with fixed spacing
          if False, elements are spaced evenly along a grid.
        shape: x, y shape of the grid (see np.reshape).
          If no shape and the list is 1D, if np.reshape were run with (1, -1).
        align_x: {'x', 'xmin', 'xmax'}
          to perform the x (column) alignment along
        align_y: {'y', 'ymin', 'ymax'}
          to perform the y (row) alignment along
        edge_x: {'x', 'xmin', 'xmax'}
          to perform the x (column) distribution (ignored if separation = True)
        edge_y: {'y', 'ymin', 'ymax'}
          to perform the y (row) distribution along (ignored if separation = True)
        rotation: for each component in degrees
        h_mirror: horizontal mirror using y axis (x, 1) (1, 0). This is the most common mirror.
        v_mirror: vertical mirror using x axis (1, y) (0, y)

    Returns:
        Component containing all the components in a grid.
    """

    device_array = np.asarray(components)
    # Check arguments
    if device_array.ndim not in (1, 2):
        raise ValueError("[PHIDL] grid() The components needs to be 1D or 2D.")
    if shape is not None and len(shape) != 2:
        raise ValueError(
            "[PHIDL] grid() shape argument must be None or"
            + " have a length of 2, for example shape=(4,6)"
        )

    # Check that shape is valid and reshape array if needed
    if (shape is None) and (device_array.ndim == 2):  # Already in desired shape
        shape = device_array.shape
    elif (shape is None) and (device_array.ndim == 1):
        shape = (device_array.size, -1)
    elif 0 < shape[0] * shape[1] < device_array.size:
        raise ValueError(
            "[PHIDL] grid() The shape is too small for all the items in components"
        )
    else:
        if np.min(shape) == -1:
            remainder = np.max(shape) - device_array.size % np.max(shape)
        else:
            remainder = shape[0] * shape[1] - device_array.size
        if remainder != 0:
            device_array = np.append(
                device_array,
                [
                    None,
                ]
                * remainder,
            )
    device_array = np.reshape(device_array, shape)

    D = Component("grid")
    ref_array = np.empty(device_array.shape, dtype=object)
    dummy = Component()
    for idx, d in np.ndenumerate(device_array):
        if d is not None:
            d = d() if callable(d) else d
            ref = d.ref(rotation=rotation, h_mirror=h_mirror, v_mirror=v_mirror)
            D.add(ref)
            ref_array[idx] = ref

        else:
            ref_array[idx] = D << dummy  # Create dummy devices
        D.aliases[idx] = ref_array[idx]

    rows = [Group(ref_array[n, :]) for n in range(ref_array.shape[0])]
    cols = [Group(ref_array[:, n]) for n in range(ref_array.shape[1])]

    # Align rows and columns independently
    for r in rows:
        r.align(alignment=align_y)
    for c in cols:
        c.align(alignment=align_x)

    # Distribute rows and columns
    Group(cols).distribute(
        direction="x", spacing=spacing[0], separation=separation, edge=edge_x
    )
    Group(rows[::-1]).distribute(
        direction="y", spacing=spacing[1], separation=separation, edge=edge_y
    )

    return D


@cell
def grid_with_text(
    components: Tuple[ComponentOrFactory, ...],
    text_prefix: str = "",
    text_offset: Float2 = (0, 0),
    text: ComponentFactory = text_rectangular,
    **kwargs,
) -> Component:
    """Returns Grid with text labels

    Args:
        text_prefix: Optional prefix
        text_offset:
        text: factory for text

    keyword Args:
        components: Iterable to be placed onto a grid. (can be 1D or 2D)
        spacing: between adjacent elements on the grid, can be a tuple for
          different distances in height and width.
        separation: If True, guarantees elements are speparated with fixed spacing
          if False, elements are spaced evenly along a grid.
        shape: x, y shape of the grid (see np.reshape).
          If no shape and the list is 1D, if np.reshape were run with (1, -1).
        align_x: {'x', 'xmin', 'xmax'}
          to perform the x (column) alignment along
        align_y: {'y', 'ymin', 'ymax'}
          to perform the y (row) alignment along
        edge_x: {'x', 'xmin', 'xmax'}
          to perform the x (column) distribution (ignored if separation = True)
        edge_y: {'y', 'ymin', 'ymax'}
          to perform the y (row) distribution along (ignored if separation = True)
        rotation: for each reference in degrees

    """
    c = Component()
    g = grid(components=components, **kwargs)
    c << g
    for i, ref in enumerate(g.aliases.values()):
        t = c << text(f"{text_prefix}{i}")
        t.move(ref.center + text_offset)
    return c


def test_grid():
    import gdsfactory as gf

    components = [gf.components.rectangle(size=(i, i)) for i in range(1, 10)]
    c = grid(components)
    difftest(c)
    return c


if __name__ == "__main__":
    import gdsfactory as gf

    # components = [gf.components.rectangle(size=(i, i)) for i in range(40, 66, 5)]
    # components = [gf.components.rectangle(size=(i, i)) for i in range(40, 66, 5)]

    components = [gf.components.triangle(x=i) for i in range(1, 10)]
    c = grid(
        components,
        shape=(1, len(components)),
        rotation=0,
        h_mirror=False,
        v_mirror=True,
    )
    c.show()



import gdsfactory.simulation.modes as gm


def test_neff_vs_width(dataframe_regression):
    df = gm.find_neff_vs_width(steps=3)
    dataframe_regression.check(df)


if __name__ == "__main__":
    test_neff_vs_width()



"""Compute group and effective index for different waveguide widths and heights.

Reproduce Yufei thesis results with MPB.

https://www.photonics.intec.ugent.be/contact/people.asp?ID=332
"""

import pathlib

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pydantic
from scipy.interpolate import interp2d

from gdsfactory.config import PATH
from gdsfactory.simulation.modes.find_mode_dispersion import find_mode_dispersion

PATH.modes = pathlib.Path.cwd() / "data"

w0 = 0.465
h0 = 0.215

dwmax = 10e-3
dhmax = 5e-3


@pydantic.validate_arguments
def find_neff_ng_dw_dh(
    ncore: float = 3.47668,
    nclad: float = 1.44401,
    dwmax: float = 30e-3,
    dhmax: float = 20e-3,
    w0: float = w0,
    h0: float = h0,
    wavelength: float = 1.55,
    steps: int = 11,
    mode_number: int = 1,
) -> pd.DataFrame:
    """Computes group and effective index for different widths and heights."""
    dw = np.linspace(-dwmax, dwmax, steps)
    dh = np.linspace(-dhmax, dhmax, steps)

    neffs = []
    ngs = []
    dhs = []
    dws = []

    for dwi in dw:
        for dhi in dh:
            m = find_mode_dispersion(
                wg_width=w0 + dwi,
                wg_thickness=h0 + dhi,
                wavelength=wavelength,
                mode_number=mode_number,
            )
            neffs.append(m.neff)
            ngs.append(m.ng)
            dws.append(dwi)
            dhs.append(dhi)

    df = pd.DataFrame(dict(dw=dws, dh=dhs, neff=neffs, ng=ngs))
    return df


def plot_neff_ng_dw_dh(
    w0: float = w0,
    h0: float = h0,
    wavelength: float = 1.55,
    mode_number: int = 1,
    **kwargs
) -> None:
    """

    Args:
        w0: center width
        h0: center height
        wavelength:
        mode_number: 1 is the fundamental first order mode

    """

    filepath = pathlib.Path(PATH.modes / "mpb_dw_dh_dispersion.csv")
    m = find_mode_dispersion(wg_width=w0, wg_thickness=h0, wavelength=wavelength)
    neff0 = m.neff
    ng0 = m.ng

    if filepath.exists():
        df = pd.read_csv(filepath)
    else:
        df = find_neff_ng_dw_dh(wavelength=wavelength, **kwargs)
        dirpath = filepath.parent
        dirpath.mkdir(exist_ok=True, parents=True)
        df.to_csv(filepath)

    dws = df.dw.values
    dhs = df.dh.values
    ngs = df.ng.values
    neffs = df.neff.values

    # neff interpolation
    f_w = interp2d(neffs, ngs, np.array(dws), kind="cubic")
    f_h = interp2d(neffs, ngs, np.array(dhs), kind="cubic")

    ws = w0 + np.array(dws)
    hs = h0 + np.array(dhs)

    plt.plot(ws * 1e3, hs * 1e3, "ko")
    extracted_dw = []
    extracted_dh = []

    for neff, ng in zip(neffs, ngs):
        temp_w = f_w(neff, ng) + w0
        temp_h = f_h(neff, ng) + h0
        extracted_dw.append(temp_w * 1e3)
        extracted_dh.append(temp_h * 1e3)

    plt.plot(extracted_dw, extracted_dh, "rx")
    plt.xlabel("width (nm)")
    plt.ylabel("height (nm)")

    plt.figure()
    plt.plot(neffs, ngs, "ro")
    plt.plot(neff0, ng0, "bx")
    plt.xlabel("neff")
    plt.ylabel("ng")
    plt.show()


if __name__ == "__main__":
    plot_neff_ng_dw_dh()



"""Merge mask metadata with test and data analysis protocols

config.yml

```yaml
test_protocols:
    passive_optical_te_coarse:
        wl_min:
        wl_max:
        wl_step:
        polarization: te

    passive_optical_tm_coarse:
        wl_min:
        wl_max:
        wl_step:
        polarization: tm
    ...

```


does.yml

```yaml
doe01:
    instances:
        - cell_name1, x1, y1
        - cell_name2, x2, y2
        - cell_name3, x3, y3

    test_protocols:
        - passive_optical_te_coarse

doe02:
    instances:
        - cell_name21, x21, y21
        - cell_name22, x22, y22
        - cell_name23, x23, y23

    test_protocols:
        - passive_optical_te_coarse
    ...
```
"""

import pathlib
from pathlib import Path
from typing import List

from omegaconf import DictConfig, OmegaConf

from gdsfactory.config import CONFIG, logger


def parse_csv_data(csv_labels_path: Path) -> List[List[str]]:
    """Returns CSV labels."""
    with open(csv_labels_path) as f:
        # Get all lines
        lines = [line.replace("\n", "") for line in f.readlines()]

        # Ignore labels for metrology structures
        lines = [line for line in lines if not line.startswith("METR_")]

        # Split lines in fields
        lines = [line.split(",") for line in lines]

        lines = [[s.strip() for s in splitted if s.strip()] for splitted in lines]

        # Remove empty lines
        lines = [line for line in lines if line]
    return lines


def get_cell_from_label(label: str) -> str:
    """get cell name from the label (cell_name is in parenthesis)"""
    cell_name = label.split("(")[1].split(")")[0]
    if cell_name.startswith("loopback"):
        cell_name = "_".join(cell_name.split("_")[1:])
    return cell_name


def merge_test_metadata(
    gdspath: Path = CONFIG["mask_gds"], labels_prefix: str = "opt"
) -> DictConfig:
    """Returns a test metadata dict config of labeled cells

    Args:
        gdspath
        labels_prefix

    """
    gdspath = pathlib.Path(gdspath)
    mask_metadata_path = gdspath.with_suffix(".yml")
    csv_labels_path = gdspath.with_suffix(".csv")
    test_metadata_path = gdspath.with_suffix(".tp.yml")

    if not mask_metadata_path.exists():
        raise FileNotFoundError(f"missing mask YAML metadata {mask_metadata_path}")

    if not csv_labels_path.exists():
        raise FileNotFoundError(f"missing CSV labels {csv_labels_path}")

    labels_list = parse_csv_data(csv_labels_path)
    metadata = OmegaConf.load(mask_metadata_path)
    cells_metadata = metadata.get("cells", {})

    test_metadata = DictConfig({})

    for label, x, y in labels_list:
        cell = get_cell_from_label(label)

        if cell in cells_metadata:
            test_metadata[cell] = cells_metadata[cell]
            test_metadata[cell].label = dict(x=x, y=y, text=label)
        else:
            logger.error(f"missing cell metadata for {cell}")

    OmegaConf.save(test_metadata, f=test_metadata_path)
    return test_metadata


if __name__ == "__main__":
    from gdsfactory import CONFIG

    # gdspath = CONFIG["repo_path"] / "samples" / "mask" / "build" / "mask" / "mask.gds"
    gdspath = (
        CONFIG["samples_path"] / "mask_pack" / "build" / "mask" / "sample_mask.gds"
    )
    d = merge_test_metadata(gdspath)
    print(d)
